package ru.dbotthepony.kstarbound.client.render

import org.lwjgl.opengl.GL11.GL_COLOR_BUFFER_BIT
import org.lwjgl.opengl.GL11.GL_RGBA
import org.lwjgl.opengl.GL11.glClear
import ru.dbotthepony.kstarbound.client.gl.BlendFunc
import ru.dbotthepony.kstarbound.client.gl.GLFrameBuffer
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.GLTexture2D
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kstarbound.client.gl.program.GLHardLightGeometryProgram
import ru.dbotthepony.kstarbound.client.gl.program.GLSoftLightGeometryProgram
import ru.dbotthepony.kstarbound.client.gl.vertex.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.quad
import ru.dbotthepony.kvector.matrix.Matrix4fStack
import ru.dbotthepony.kvector.vector.Color
import ru.dbotthepony.kvector.vector.nfloat.Vector2f
import ru.dbotthepony.kvector.vector.nfloat.Vector3f

// Huge thanks to articles by Scott [slembcke] Lembcke!
// https://slembcke.github.io/SuperFastHardShadows
// https://slembcke.github.io/SuperFastSoftShadows
class LightRenderer(val state: GLStateTracker) {
	interface ShadowGeometryRenderer {
		fun renderHardGeometry(renderer: LightRenderer, lightPosition: Vector2f, lightRadius: Float, stack: Matrix4fStack, program: GLHardLightGeometryProgram)
		fun renderSoftGeometry(renderer: LightRenderer, lightPosition: Vector2f, lightRadius: Float, stack: Matrix4fStack, program: GLSoftLightGeometryProgram)
	}

	private val geometry = ArrayList<ShadowGeometryRenderer>()

	fun addShadowGeometry(geometry: ShadowGeometryRenderer): LightRenderer {
		this.geometry.add(geometry)
		return this
	}

	fun removeShadowGeometry(geometry: ShadowGeometryRenderer): Boolean {
		return this.geometry.remove(geometry)
	}

	fun clearShadowGeometry() {
		geometry.clear()
	}

	/**
	 * Сюда происходит рендер маски света и самого света
	 */
	private val framebufferRender = GLFrameBuffer(state)

	/**
	 * Сюда накапливается отрисованный свет
	 */
	val framebufferAccumulator = GLFrameBuffer(state)

	val outputTexture: GLTexture2D? get() = framebufferAccumulator.texture

	fun resizeFramebuffer(width: Int, height: Int) {
		framebufferRender.reattachTexture(width, height, GL_RGBA)
		framebufferAccumulator.reattachTexture(width, height, GL_RGBA)
	}

	fun begin(clearGeometry: Boolean = true) {
		state.ensureSameThread()

		if (!framebufferRender.isComplete || !framebufferAccumulator.isComplete) {
			return
		}

		val old = state.clearColor
		state.clearColor = Color.BLACK

		framebufferRender.bind()
		glClear(GL_COLOR_BUFFER_BIT)
		checkForGLError()
		framebufferAccumulator.bind()
		glClear(GL_COLOR_BUFFER_BIT)
		checkForGLError()
		state.clearColor = old

		framebufferAccumulator.unbind()

		if (clearGeometry) {
			geometry.clear()
		}
	}

	/**
	 * Отрисовывает результат на весь текущий framebuffer в режиме additive
	 */
	fun renderOutputAdditive() {
		val oldFunc = state.blendFunc

		try {
			state.activeTexture = 0
			state.texture2D = outputTexture
			state.blendFunc = BlendFunc.ADDITIVE
			state.programs.textureQuad.run(0)
		} finally {
			state.blendFunc = oldFunc
		}
	}

	/**
	 * Отрисовывает результат на весь текущий framebuffer в режиме additive
	 */
	fun renderOutputMultiplicative() {
		val oldFunc = state.blendFunc

		try {
			state.activeTexture = 0
			state.texture2D = outputTexture
			state.blendFunc = BlendFunc.MULTIPLY_WITH_ALPHA
			state.programs.textureQuad.run(0)
		} finally {
			state.blendFunc = oldFunc
		}
	}

	private fun blendResult() {
		val oldFunc = state.blendFunc

		try {
			framebufferAccumulator.bind()

			state.blendFunc = BlendFunc.ADDITIVE
			state.activeTexture = 0
			state.texture2D = framebufferRender.texture
			state.programs.textureQuad.run(0)
		} finally {
			state.blendFunc = oldFunc
			framebufferAccumulator.unbind()
		}
	}

	fun renderHardLight(
		position: Vector2f,
		color: Color = Color.WHITE,
		radius: Float = 10f,
		stack: Matrix4fStack = state.matrixStack
	) {
		state.ensureSameThread()

		if (!framebufferRender.isComplete || !framebufferAccumulator.isComplete) {
			return
		}

		val oldFunc = state.blendFunc

		// отрисовка
		try {
			framebufferRender.bind()

			// state.programs.colorQuad.clearAlpha()

			// Геометрия
			val old = state.clearColor
			state.clearColor = Color.BLACK
			glClear(GL_COLOR_BUFFER_BIT)
			checkForGLError()
			state.clearColor = old

			state.programs.hardLightGeometry.use()
			state.programs.hardLightGeometry.transform.set(stack.last)
			state.programs.hardLightGeometry.lightPosition.set(position)

			state.blendFunc = BlendFunc.ONLY_ALPHA

			for (renderer in geometry) {
				renderer.renderHardGeometry(this, position, radius, stack, state.programs.hardLightGeometry)
			}

			state.programs.light.use()
			state.programs.light.transform.set(stack.last)
			state.programs.light.baselineColor.set(color)

			// Свет
			val builder = state.programs.light.builder

			builder.begin()
			builder.quad(position.x - radius, position.y - radius, position.x + radius, position.y + radius, QuadTransformers.uv())
			builder.upload()

			state.blendFunc = BLEND_MODE
			builder.draw()
		} finally {
			state.blendFunc = oldFunc
			framebufferRender.unbind()
		}

		blendResult()
	}

	fun renderSoftLight(
		position: Vector2f,
		color: Color = Color.WHITE,
		radius: Float = 10f,
		innerRadius: Float = radius / 3f,
		stack: Matrix4fStack = state.matrixStack
	) {
		state.ensureSameThread()

		if (!framebufferRender.isComplete || !framebufferAccumulator.isComplete) {
			return
		}

		val oldFunc = state.blendFunc

		// отрисовка
		try {
			framebufferRender.bind()

			// state.programs.colorQuad.clearAlpha()

			// Геометрия
			val old = state.clearColor
			state.clearColor = CLEAR_COLOR_SOFT
			glClear(GL_COLOR_BUFFER_BIT)
			checkForGLError()
			state.clearColor = old

			state.programs.softLightGeometry.use()
			state.programs.softLightGeometry.transform.set(stack.last)
			state.programs.softLightGeometry.lightPositionAndSize.set(Vector3f(position, innerRadius))
			state.programs.softLightGeometry.lightPenetration.set(1f)

			state.blendFunc = BlendFunc.ONLY_BLEND_ALPHA

			for (renderer in geometry) {
				renderer.renderSoftGeometry(this, position, radius, stack, state.programs.softLightGeometry)
			}

			state.programs.light.use()
			state.programs.light.transform.set(stack.last)
			state.programs.light.baselineColor.set(color)

			// Свет
			val builder = state.programs.light.builder

			builder.begin()
			builder.quad(position.x - radius, position.y - radius, position.x + radius, position.y + radius, QuadTransformers.uv())
			builder.upload()

			state.blendFunc = BLEND_MODE_SOFT
			builder.draw()
		} finally {
			state.blendFunc = oldFunc
			framebufferRender.unbind()
		}

		blendResult()
	}

	val builder by lazy {
		StreamVertexBuilder(state, SHADOW_FORMAT, GeometryType.QUADS_AS_LINES, 256)
	}

	val hugeBuilder by lazy {
		StreamVertexBuilder(state, SHADOW_FORMAT, GeometryType.QUADS_AS_LINES, 16384)
	}

	val lineBuilder by lazy {
		StreamVertexBuilder(state, SHADOW_FORMAT, GeometryType.LINES, 256)
	}

	val builderSoft by lazy {
		StreamVertexBuilder(state, SHADOW_FORMAT_SOFT, GeometryType.QUADS_ALTERNATIVE, 256)
	}

	val lineBuilderSoft by lazy {
		StreamVertexBuilder(state, SHADOW_FORMAT_SOFT, GeometryType.LINES, 256)
	}

	companion object {
		val CLEAR_COLOR_SOFT = Color(0f, 0f, 0f, 0f)

		val BLEND_MODE = BlendFunc(
			BlendFunc.Func.DST_ALPHA,
			BlendFunc.Func.ZERO,
			BlendFunc.Func.ZERO,
			BlendFunc.Func.ONE,
		)

		val BLEND_MODE_SOFT = BlendFunc(
			BlendFunc.Func.ONE_MINUS_DST_ALPHA,
			BlendFunc.Func.ZERO,
			BlendFunc.Func.ZERO,
			BlendFunc.Func.ONE,
		)

		val SHADOW_FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
		val SHADOW_FORMAT_SOFT = GLAttributeList.Builder().push(GLType.VEC4F).push(GLType.VEC2F).build()
	}
}
